# ============================================================== #
# This is open data under the CC BY license
# http://creativecommons.org/licenses/by/4.0/
# Python code to train CNN for combustion regime identification  #
# with Raman/Rayleigh line measurements data of MRB flames.      #
#                                                                #
# K.D. Wan, S. Hartl, L. Vervisch, P. Domingo, R.S. Barlow,      #
# C. Hasse, Combustion regime identification from machine        #
# learning trained by Raman/Rayleigh line measurements,          #
# Combust. Flame 219: 268-274, 2020                              #
# ============================================================== #

import tensorflow as tf
import numpy as np
import matplotlib.pyplot as plt
import time
from sklearn.utils import shuffle
from tensorflow.keras.layers import Input, Concatenate, Conv1D, MaxPooling1D, UpSampling1D, Flatten, Dense, Dropout, concatenate
from tensorflow.keras.models import Model

print(tf.__version__)

# Load the Raman/Rayleigh line measurements data of MRB flames
train_database = []
test_database = []

file_names = [ \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_015_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_022_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_030_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_060_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_090_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_015_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_022_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_030_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_060_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_090_GFRI_Exp.npy', \
]

for ifile in range(len(file_names)):
    train_database.extend(list(np.load(file_names[ifile])))

#####################################

file_names = [ \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_015_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_022_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_030_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_060_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB18b_final_090_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_015_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_022_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_030_GFRI_Exp.npy', \
'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_060_GFRI_Exp.npy', \
# 'EXPORT_NEW_DATA_256/26-Feb-2020_MRB26b_final_090_GFRI_Exp.npy', \
]

for ifile in range(len(file_names)):
    test_database.extend(list(np.load(file_names[ifile])))

#####################################

train_database = np.array(train_database)
test_database = np.array(test_database)

col_rm = [8,9,10,11,12,13,14,15,17] # Keep only species mass fraction and temperature
train_database = np.delete(train_database,col_rm,axis=-1)
test_database = np.delete(test_database,col_rm,axis=-1)

train_labels = train_database[:,:,-1]
train_input = train_database[:,:,0:-1]
test_labels = test_database[:,:,-1]
test_input = test_database[:,:,0:-1]

#############################################
nvar = len(train_input[0,0])
input_max = np.zeros(nvar)

# Normalization
input_max = [0.205, 0.233, 0.329, 0.767,
             0.074, 0.235, 0.005, 2293.3]

for n in range(nvar):
    train_input[:,:,n] = train_input[:,:,n]/input_max[n]
    test_input[:,:,n] = test_input[:,:,n]/input_max[n]

train_label = np.expand_dims(train_labels,axis=-1)
test_label = np.expand_dims(test_labels,axis=-1)

# Adjust the values of the combustion regime index \xi
train_label[train_label==4] = 3
train_label[train_label==5] = 4
train_label[train_label==6] = 5
train_label[train_label==7] = 6

test_label[test_label==4] = 3
test_label[test_label==5] = 4
test_label[test_label==6] = 5
test_label[test_label==7] = 6

train_label[train_label==-1] = 7 # Dummy value
test_label[test_label==-1] = 7

# Data augmentation
#############################################
# Mirror
train_input_R = train_input[:,::-1,:]
train_label_R = train_label[:,::-1,:]

train_input1 = np.concatenate((train_input, train_input_R), axis=0)
train_label1 = np.concatenate((train_label, train_label_R), axis=0)

#############################################
# Noise
train_input_N = np.random.rand(train_input1.shape[0],train_input1.shape[1],train_input1.shape[2])
train_input_N = (train_input_N*2-1)*5/100+1
train_input_N = train_input1*train_input_N

train_input2 = np.concatenate((train_input1, train_input_N), axis=0)
train_label2 = np.concatenate((train_label1, train_label1), axis=0)

#############################################
train_input2, train_label2 = shuffle(train_input2, train_label2)
#############################################

# Build the CNN
input_layer = Input((256,nvar))
conv1 = Conv1D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(input_layer)
conv1 = Conv1D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv1)
pool1 = MaxPooling1D(pool_size=2)(conv1)
conv2 = Conv1D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool1)
conv2 = Conv1D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv2)
pool2 = MaxPooling1D(pool_size=2)(conv2)
conv3 = Conv1D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool2)
conv3 = Conv1D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv3)
pool3 = MaxPooling1D(pool_size=2)(conv3)
conv4 = Conv1D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool3)
conv4 = Conv1D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv4)
drop4 = Dropout(0.5)(conv4)
pool4 = MaxPooling1D(pool_size=2)(drop4)

conv5 = Conv1D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(pool4)
conv5 = Conv1D(1024, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv5)
drop5 = Dropout(0.5)(conv5)

up6 = Conv1D(512, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling1D(size = 2)(drop5))
merge6 = concatenate([drop4,up6], axis = 2)
conv6 = Conv1D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge6)
conv6 = Conv1D(512, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv6)

up7 = Conv1D(256, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling1D(size = 2)(conv6))
merge7 = concatenate([conv3,up7], axis = 2)
conv7 = Conv1D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge7)
conv7 = Conv1D(256, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv7)

up8 = Conv1D(128, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling1D(size = 2)(conv7))
merge8 = concatenate([conv2,up8], axis = 2)
conv8 = Conv1D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge8)
conv8 = Conv1D(128, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv8)

up9 = Conv1D(64, 2, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(UpSampling1D(size = 2)(conv8))
merge9 = concatenate([conv1,up9], axis = 2)
conv9 = Conv1D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(merge9)
conv9 = Conv1D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
conv9 = Conv1D(64, 3, activation = 'relu', padding = 'same', kernel_initializer = 'he_normal')(conv9)
conv10 = Conv1D(8, 1, activation = 'softmax')(conv9)

model = Model(inputs = input_layer, outputs = conv10)

model.compile(optimizer=tf.keras.optimizers.Adam(lr=3e-4,decay=1e-4),
              loss='sparse_categorical_crossentropy',
              metrics=['accuracy'])

# CNN training
start = time.time()
start_cpu = time.clock()
model.fit(train_input2, train_label2, batch_size=1024, epochs=200)
end = time.time()
end_cpu = time.clock()
print("Run time: %.1fs, CPU time: %.1fs" % (end-start,end_cpu-start_cpu))

# CNN testing
print("Evaluating the model:")
loss,accuracy = model.evaluate(test_input, test_label, batch_size=1024)

res = model.predict(test_input, batch_size=1024)
ref = test_label

# Get the regime index from the CNN output
ress = np.zeros(ref.shape)
for i in range(len(ref)):
    for j in range(len(ref[0])):
        ress[i][j] = np.argmax(res[i,j])

res = ress

# Compute the pixel-wise accuracy
error = 0
valid = 0
for i in range(len(ref)):
    for j in range(len(ref[0])):
        if (ref[i][j] != 7):
            valid = valid+1
            if (ref[i][j]!=res[i][j]):
                error = error+1

1.0-error/valid
